
using System;
using System.Collections;
using System.Reflection;
using System.Threading;
using System.IO;
using Mono.Unix;
using Gtk;
using IPod;

namespace Dopi {

    public class DopiWindow : Window {

        private Button updateButton;
        private Button refreshButton;
        private Button ejectButton;
        private DeviceCombo combo;
        private SongView view;
        private ProgressDialog pd;
        private UIManager ui;
        private ActionGroup actions;
        private ActionGroup playlistActions;
        private ThreadNotify notify;

        private PlaylistWindow pw;
        
        private Device device;

        private bool updating = false;
        private bool busy = false;

        public Device Device {
            get { return device; }
        }
        
        public bool IsBusy {
            get { return busy; }
            set {
                busy = value;

                Application.Invoke (delegate {
                    if (busy) {
                        Gdk.Cursor cursor = new Gdk.Cursor (Gdk.CursorType.Watch);
                        this.GdkWindow.Cursor = cursor;
                    } else {
                        this.GdkWindow.Cursor = null;
                    }
                });
            }
        }

        public DopiWindow () : base (Catalog.GetString ("Dopi - iPod Update Tool")) {

            Gdk.Pixbuf pixbuf = IconTheme.Default.LoadIcon ("gnome-dev-ipod", 24, 0);
            if (pixbuf != null) {
                this.Icon = pixbuf;
            }

            notify = new ThreadNotify (new ReadyEvent (OnNotify));
            DefaultHeight = 400;
            VBox box = new VBox (false, 6);
            box.BorderWidth = 6;

            HBox hbox = new HBox (false, 6);

            ejectButton = new Button (new Image (Assembly.GetExecutingAssembly (), "media-eject.png"));
            ejectButton.Clicked += OnEject;
            hbox.PackStart (ejectButton, false, false, 0);
            
            combo = new DeviceCombo ();
            combo.Changed += OnDeviceChanged;
            hbox.PackStart (combo, true, true, 0);

            refreshButton = new Button (Stock.Refresh);
            refreshButton.Clicked += OnRefresh;
            hbox.PackStart (refreshButton, false, false, 0);

            updateButton = Utility.CreateButton (Stock.Save, Catalog.GetString ("_Update iPod"));
            updateButton.Clicked += OnUpdate;
            hbox.PackStart (updateButton, false, false, 0);

            box.PackStart (hbox, false, false, 0);
            
            ScrolledWindow scroll = new ScrolledWindow ();
            scroll.HscrollbarPolicy = PolicyType.Automatic;
            scroll.VscrollbarPolicy = PolicyType.Automatic;
            scroll.ShadowType = ShadowType.EtchedIn;

            box.PackStart (scroll, true, true, 0);
            
            view = new SongView (this);
            view.SongEdited += OnSongEdited;
            scroll.Add (view);

            box.ShowAll ();
            Add (box);

            this.DeleteEvent += OnDeleteEvent;

            pd = new ProgressDialog (this);
            pd.Modal = true;

            ActionEntry[] actionList = {
                new ActionEntry ("AddFile", Stock.Open,
                                 Catalog.GetString ("Add _Files..."), "<control>a",
                                 Catalog.GetString ("Add files to the iPod"),
                                 OnAddFile),
                new ActionEntry ("AddDirectory", Stock.Open,
                                 Catalog.GetString ("Add _Directory..."), "<shift><control>a",
                                 Catalog.GetString ("Add a directory to the iPod"),
                                 OnAddDirectory),
                new ActionEntry ("Playlists", Stock.Properties,
                                 Catalog.GetString ("Edit Playlists..."), "<control>p",
                                 Catalog.GetString ("Edit the playlists on the iPod"),
                                 OnEditPlaylists),
                new ActionEntry ("NewPlaylist", null,
                                 Catalog.GetString ("New Playlist..."), null, null, null),
                new ActionEntry ("Remove", Stock.Delete, Catalog.GetString ("_Remove Song(s)"), null,
                                 Catalog.GetString ("Remove a song from the iPod"),
                                 OnRemoveSong),
                new ActionEntry ("RemoveDups", Stock.Delete, Catalog.GetString ("_Remove Duplicates"), null,
                                 Catalog.GetString ("Remove duplicate songs from the iPod"),
                                 OnRemoveDups),
                new ActionEntry ("Properties", Stock.Properties, Catalog.GetString ("Device _Properties..."), null,
                                 Catalog.GetString ("Show the properties for the current device"),
                                 OnProperties)
            };

            actions = new ActionGroup ("Dopi Actions");
            actions.Add (actionList);
            
            ui = new UIManager ();
            ui.InsertActionGroup (actions, 0);
            ui.AddUiFromResource ("dopi.xml");

            playlistActions = new ActionGroup ("Playlist Actions");
            ui.InsertActionGroup (playlistActions, 0);

            view.ButtonPressEvent += OnButtonPress;

            pw = new PlaylistWindow ();

            SetDevice ();
        }

        private void SetDevice (Device device) {

            if (device != null) {
                Console.WriteLine ("Changing to device '{0}'", device.Name);
                this.Sensitive = true;
                view.SongDatabase = device.SongDatabase;
                pd.SongDatabase = device.SongDatabase;
                pw.SongDatabase = device.SongDatabase;
            } else {
                Console.WriteLine ("No device");
                this.Sensitive = false;
                view.SongDatabase = null;
                pd.SongDatabase = null;
                pw.SongDatabase = null;
            }

            this.device = device;
            
            view.Refresh ();
        }

        private void SetDevice () {
            try {
                SetDevice (combo.ActiveDevice);
            } catch (Exception e) {
                ShowError (Catalog.GetString ("There was a problem while changing devices:"),
                           e.Message);
                Console.Error.WriteLine (e);
            }
        }

        private void OnSongEdited (object o, Song song) {
            pw.RefreshSong (song);
        }

        private void OnDeviceChanged (object o, EventArgs args) {
            SetDevice ();
        }

        private void OnUpdate (object o, EventArgs args) {
            SetSensitive (false);
            IsBusy = true;
            Thread thread = new Thread (new ThreadStart (DoUpdate));
            thread.Start ();
        }
        
        private void OnRefresh (object o, EventArgs args) {
            IsBusy = true;
            while (Application.EventsPending ())
                Application.RunIteration ();

            try {
                view.SongDatabase.Reload ();
                view.Refresh ();
                pw.Refresh ();
            } catch (Exception e) {
                ShowError (Catalog.GetString ("There was a problem refreshing:"), e.Message);
            } finally {
                IsBusy = false;
            }
        }

        private void OnEject (object o, EventArgs args) {
            try {
                device.Eject ();
            } catch (Exception e) {
                ShowError (Catalog.GetString ("There was a problem ejecting the device:"), e.Message);
            }
        }

        [GLib.ConnectBefore]
        private void OnButtonPress (object o, ButtonPressEventArgs args) {
            if (args.Event.Button == 3) {
                Menu menu = (Menu) ui.GetWidget ("/ui/DopiPopup");
                menu.Popup ();
                args.RetVal = true;
            }
        }

        private void DoUpdate () {
            try {
                device.UserName = Environment.UserName;
                device.HostName = Environment.MachineName;
                device.Save ();
                
                lock (view.SongDatabase) {
                    updating = true;
                    view.SongDatabase.Save ();
                    notify.WakeupMain ();
                }
            } catch (Exception e) {
                Console.WriteLine ("Error saving iPod: " + e);
            } finally {
                updating = false;
            }
        }

        private void OnNotify () {
            IsBusy = false;
            SetSensitive (true);
        }

        private void OnDeleteEvent (object o, DeleteEventArgs args) {
            if (updating) {
                args.RetVal = true;
            } else { 
                Application.Quit ();
            }
        }

        private void OnAddFile (object o, EventArgs args) {
            FileChooserDialog dialog = new FileChooserDialog (Catalog.GetString ("Add Files..."), this,
                                                              FileChooserAction.Open,
                                                              Stock.Cancel, ResponseType.Cancel,
                                                              Stock.Open, ResponseType.Ok);
            dialog.DefaultResponse = ResponseType.Ok;
            dialog.LocalOnly = true;
            dialog.SelectMultiple = true;

            FileFilter songFilter = new FileFilter ();
            songFilter.Name = Catalog.GetString ("Songs");
            songFilter.AddCustom (FileFilterFlags.Filename, OnFilterFilename);
            dialog.AddFilter (songFilter);

            ResponseType response = (ResponseType) dialog.Run ();
            string[] uris = dialog.Uris;
            dialog.Destroy ();
            
            if (response == ResponseType.Ok) {
                foreach (string uristr in uris) {
                    Uri uri = new Uri (uristr);
                    view.AddSong (uri.LocalPath);
                }
            }
        }

        private bool OnFilterFilename (FileFilterInfo info) {
            return info.MimeType == "audio/mpeg" ||
                info.MimeType == "audio/x-m4a" ||
                view.IsPlayable (info.Filename);
        }

        private void OnAddDirectory (object o, EventArgs args) {
            FileChooserDialog dialog = new FileChooserDialog (Catalog.GetString ("Add Directory..."), this,
                                                              FileChooserAction.SelectFolder,
                                                              Stock.Cancel, ResponseType.Cancel,
                                                              Stock.Open, ResponseType.Ok);
            dialog.DefaultResponse = ResponseType.Ok;
            dialog.LocalOnly = true;
            dialog.SelectMultiple = true;

            ResponseType response = (ResponseType) dialog.Run ();
            string[] uris = dialog.Uris;
            dialog.Destroy ();
            
            if (response == ResponseType.Ok) {
                foreach (string uristr in uris) {
                    Uri uri = new Uri (uristr);
                    view.AddDirectory (uri.LocalPath);
                }
            }
        }

        private void OnEditPlaylists (object o, EventArgs args) {
            pw.Show ();
        }

        private void OnRemoveSong (object o, EventArgs args) {
            view.RemoveSelected ();
        }

        private void RemoveDups (object o) {
            Song[] songs = view.GetDuplicates ();

            Application.Invoke (delegate {
                view.RemoveSongs (songs);
                IsBusy = false;
                SetSensitive (true);
            });
        }

        private void OnRemoveDups (object o, EventArgs args) {
            IsBusy = true;
            SetSensitive (false);
            ThreadPool.QueueUserWorkItem (RemoveDups);
        }

        private void OnProperties (object o, EventArgs args) {
            if (device != null) {
                PropertiesDialog dialog = new PropertiesDialog (this, device);
                dialog.Run ();
                dialog.Destroy ();
            }
        }

        public void SetSensitive (bool sensitive) {
            updateButton.Sensitive = sensitive;
            refreshButton.Sensitive = sensitive;
            ejectButton.Sensitive = sensitive;
            combo.Sensitive = sensitive;
            view.Sensitive = sensitive;
        }

        public void ShowError (string m1, string m2) {
            string message = String.Format ("<b>{0}</b>\n\n{1}", GLib.Markup.EscapeText (m1), m2);

            MessageDialog dialog = new MessageDialog (this, DialogFlags.Modal, MessageType.Error,
                                                      ButtonsType.Ok, message);
            dialog.Title = Catalog.GetString ("Dopi Error");
            dialog.Run ();
            dialog.Destroy ();
        }
    }
}
