
using System;
using System.Collections;
using Gtk;
using IPod;

namespace Dopi {

    public class PlaylistCombo : ComboBoxEntry {

        private ListStore store;
        private SongDatabase db;
        private Playlist currentPlaylist;
        private TreeIter currentIter;

        public new event EventHandler Changed;
        
        public Playlist ActivePlaylist {
            get {
                return currentPlaylist;
            } set {
                TreeIter iter = FindPlaylist (value);

                if (iter.Equals (TreeIter.Zero))
                    throw new ApplicationException (String.Format ("Unknown playlist '{0}'", value.Name));

                SetActiveIter (iter);
            }
        }

        public SongDatabase SongDatabase {
            get { return db; }
            set {
                if (db != null) {
                    db.PlaylistAdded -= OnPlaylistAdded;
                    db.PlaylistRemoved -= OnPlaylistRemoved;
                }
                
                db = value;

                if (db != null) {
                    db.PlaylistAdded += OnPlaylistAdded;
                    db.PlaylistRemoved += OnPlaylistRemoved;
                }
                
                Refresh ();
            }
        }

        public PlaylistCombo (SongDatabase db) : base () {
            this.db = db;
            store = new ListStore (typeof (string), typeof (Playlist));
            this.Model = store;

            this.TextColumn = 0;

            (Child as Entry).FocusOutEvent += OnEntryUnfocused;

            Refresh ();
        }

        public void Refresh () {
            string current = null;
            bool haveActive = false;
            
            if (ActivePlaylist != null)
                current = ActivePlaylist.Name;

            store.Clear ();

            if (db == null) {
                SetActive ();
                return;
            }

            foreach (Playlist playlist in db.Playlists) {
                TreeIter iter = store.AppendValues (playlist.Name, playlist);
                
                if (playlist.Name == current) {
                    SetActiveIter (iter);
                    haveActive = true;
                }
            }

            if (!haveActive) {
                SetActive ();
            }
        }

        private void SetActive () {
            if (store.IterNChildren () == 0) {
                store.AppendValues ("No Playlists", null);
                Sensitive = false;
                this.Active = 0;
            } else {
                Sensitive = true;
                this.Active = 0;
            }
        }

        private TreeIter FindPlaylist (Playlist pl) {
            TreeIter iter = TreeIter.Zero;

            if (!store.GetIterFirst (out iter))
                return TreeIter.Zero;

            do {
                Playlist pl2 = (Playlist) store.GetValue (iter, 1);

                if (pl == pl2)
                    return iter;
            } while (store.IterNext (ref iter));

            return TreeIter.Zero;
        }

        private void OnPlaylistAdded (object o, Playlist pl) {
            if (!Sensitive) {
                store.Clear ();
                Sensitive = true;
            }
            
            store.AppendValues (pl.Name, pl);
        }

        private void OnPlaylistRemoved (object o, Playlist pl) {
            TreeIter iter = FindPlaylist (pl);
            if (!iter.Equals (TreeIter.Zero)) {
                store.Remove (ref iter);
                SetActive ();
            }
        }

        private void OnEntryUnfocused (object o, FocusOutEventArgs args) {
            if (currentPlaylist != null) {
                string text = (Child as Entry).Text;

                currentPlaylist.Name = text;

                store.SetValue (currentIter, 0, text);
            }
        }

        private Playlist GetCurrentPlaylist () {
            if (db == null)
                return null;

            TreeIter iter = TreeIter.Zero;

            if (!GetActiveIter (out iter))
                return null;

            return (Playlist) store.GetValue (iter, 1);
        }

        protected override void OnChanged () {
            if ((Child as Widget).IsFocus)
                return;

            Playlist playlist = GetCurrentPlaylist ();

            if (playlist != currentPlaylist) {
                currentPlaylist = playlist;
                GetActiveIter (out currentIter);

                if (Changed != null)
                    Changed (this, new EventArgs ());
            }
        }
    }
}
