
using System;
using System.Threading;
using System.Collections;
using System.IO;
using System.Text;
using Mono.Unix;
using Gtk;
using IPod;
using Entagged;

namespace Dopi {

    internal enum NodeType {
        Song,
        Album,
        Artist
    }
    
    internal abstract class Node {

        private NodeType type;
        private bool marked;

        public NodeType Type {
            get { return type; }
        }

        public abstract string Title {
            get; set;
        }

        public bool Marked {
            get { return marked; }
            set { marked = value; }
        }

        public Node (NodeType type) {
            this.type = type;
            this.marked = false;
        }
    }

    internal class GenericNode : Node {

        private string title;

        public override string Title {
            get { return title; }
            set { title = value; }
        }

        public GenericNode (NodeType type, string title) : base (type) {
            this.title = title;
        }
    }

    internal class SongNode : Node {

        private Song song;
        
        public Song Song {
            get { return song; }
            set {
                song = value;
                Title = song.Title;
            }
        }

        public override string Title {
            get { return song.Title; }
            set { song.Title = value; }
        }

        public SongNode (Song song) : base (NodeType.Song) {
            this.song = song;
        }
    }

    public delegate void SongHandler (object o, Song song);
    
    public class SongView : TreeView {

        private TreeStore store;
        private SongDatabase db;
        private ThreadNotify notify;
        private DopiWindow window;

        public event SongHandler SongEdited;
        
        public SongDatabase SongDatabase {
            get { return db; }
            set {
                if (db != null) {
                    db.SaveEnded -= OnSaveEnded;
                }
                
                db = value;

                if (db != null) {
                    db.SaveEnded += OnSaveEnded;
                }
            }
        }
        
        public SongView (DopiWindow window) {
            this.window = window;
            this.HeadersVisible = false;
            this.Selection.Mode = SelectionMode.Multiple;
            
            store = new TreeStore (typeof (SongNode));
            store.SetSortColumnId (0, SortType.Ascending);
            store.SetSortFunc (0, new TreeIterCompareFunc (OnTreeSort));

            this.Model = store;

            CellRendererText cell = new CellRendererText ();
            cell.Editable = true;
            cell.Edited += OnEdit;
            AppendColumn ("Songs", cell,
                          new TreeCellDataFunc (OnCellData));
            this.KeyPressEvent += OnKeyPress;

            EnableModelDragDest (new TargetEntry[] { new TargetEntry ("text/uri-list", 0, 0) },
                                 Gdk.DragAction.Copy | Gdk.DragAction.Default);

            EnableModelDragSource (Gdk.ModifierType.Button1Mask,
                                   new TargetEntry[] { new TargetEntry ("text/uri-list", 0, 0),
                                                       new TargetEntry (Utility.SongDndTarget, 0, Utility.SongDndId) },
                                   Gdk.DragAction.Copy | Gdk.DragAction.Default);
                                   

            this.DragDataGet += OnDragDataGet;
            this.DragDataReceived += OnDragDataReceived;

            this.EnableSearch = true;
            this.SearchColumn = 0;
            this.SearchEqualFunc = OnSearchEqual;

            notify = new ThreadNotify (new ReadyEvent (OnNotify));
        }

        private bool OnSearchEqual (TreeModel model, int column, string key, TreeIter iter) {
            Node node = (Node) store.GetValue (iter, column);

            return !node.Title.ToLower ().StartsWith (key.ToLower ());
        }

        private void OnKeyPress (object o, KeyPressEventArgs args) {
            if (args.Event.Key == Gdk.Key.Delete) {
                RemoveSelected ();
            }
        }

        private void GetSongs (TreeIter iter, ArrayList list) {
            Node node = (Node) store.GetValue (iter, 0);

            if (node.Type == NodeType.Song) {
                Song song = (node as SongNode).Song;
                list.Add (song);
            } else {
                TreeIter child = TreeIter.Zero;

                if (!store.IterChildren (out child, iter))
                    return;

                do {
                    GetSongs (child, list);
                } while (store.IterNext (ref child));
            }
        }

        private Song[] GetSelectedSongs () {
            ArrayList list = new ArrayList ();
            TreePath[] selection = this.Selection.GetSelectedRows ();

            if (selection == null || selection.Length == 0)
                return new Song[0];

            foreach (TreePath path in selection) {
                TreeIter iter = TreeIter.Zero;
                store.GetIter (out iter, path);

                GetSongs (iter, list);
            }

            return (Song[]) list.ToArray (typeof (Song));
        }

        private void OnDragDataGet (object o, DragDataGetArgs args) {

            Song[] songs = GetSelectedSongs ();
            StringBuilder payload = new StringBuilder ();
                
            if (args.SelectionData.Target == "text/uri-list") {
                foreach (Song song in songs) {
                    payload.Append (song.FileName + "\r\n");
                }

                args.SelectionData.Set (Gdk.Atom.Intern ("text/uri-list", false),
                                        8, System.Text.Encoding.UTF8.GetBytes (payload.ToString ()));
            } else {
                foreach (Song song in songs) {
                    payload.Append (song.Id + ";");
                }

                args.SelectionData.Set (Gdk.Atom.Intern (Utility.SongDndTarget, false),
                                        8, System.Text.Encoding.UTF8.GetBytes (payload.ToString ()));
            }
        }

        private void OnDragDataReceived (object sender, DragDataReceivedArgs args) {
            bool success = false;

            if (Drag.GetSourceWidget (args.Context) != this &&
                db != null && args.SelectionData.Length >=0 && args.SelectionData.Format == 8) {
                try {
                    string uris = Encoding.ASCII.GetString (args.SelectionData.Data, 0, args.SelectionData.Length);

                    foreach (string str in uris.Trim ().Split ('\n')) {
                        string uristr = str.Trim (); // remove the '\r'

                        if (uristr == String.Empty)
                            continue;

                        Uri uri = new Uri (uristr);

                        if (Directory.Exists (uri.LocalPath)) {
                            AddDirectory (uri.LocalPath);
                            success = true;
                        } else if (File.Exists (uri.LocalPath)) {
                            AddSong (uri.LocalPath);
                            success = true;
                        } else {
                            Console.WriteLine ("Not a valid path: {0}", uri.LocalPath);
                        }
                    }
                } catch (Exception e) {
                    window.ShowError (Catalog.GetString ("There was a problem while adding songs:"), e.Message);
                }
            }

            Drag.Finish (args.Context, success, false, args.Time);
        }

        public void RemoveSongs (Song[] songs) {
            foreach (Song song in songs) {
                db.RemoveSong (song);
                TreeIter iter = FindSong (song);
                store.Remove (ref iter);
            }
        }

        private void RemoveSongs (TreeIter iter) {
            Node node = (Node) store.GetValue (iter, 0);

            if (node.Type == NodeType.Song) {
                Song song = (node as SongNode).Song;
                db.RemoveSong (song);
            } else {
                TreeIter child = TreeIter.Zero;

                if (!store.IterChildren (out child, iter))
                    return;

                do {
                    RemoveSongs (child);
                } while (store.IterNext (ref child));
            }
        }

        public void RemoveSelected () {
            TreePath[] selection = this.Selection.GetSelectedRows ();

            if (selection != null && selection.Length > 0) {
                TreeIter iter = TreeIter.Zero;
                store.GetIter (out iter, selection[0]);
                RemoveSongs (iter);
                store.Remove (ref iter);
                RemoveSelected ();
            }
        }

        public Song[] GetDuplicates () {
            ArrayList dups = new ArrayList ();

            Song[] songs = db.Songs;
            for (int i = 0; i < songs.Length; i++) {
                if (dups.Contains (i))
                    continue;
                
                Song a = songs[i];
                
                for (int j = i + 1; j < songs.Length; j++) {
                    Song b = songs[j];
                    
                    if (a.Artist == b.Artist &&
                        a.Album == b.Album &&
                        a.Title == b.Title &&
                        a.Size == b.Size) {
                        dups.Add (j);
                    }
                }
            }

            ArrayList ret = new ArrayList ();
            foreach (int item in dups) {
                ret.Add (songs[item]);
            }

            return (Song[]) ret.ToArray (typeof (Song));
        }

        private int ArtistCompare (string a, string b) {
            string a2;
            string b2;

            if (a.ToLower ().StartsWith ("the ") && a.Length > 4) {
                a2 = a.Substring (4).Trim ();
            } else {
                a2 = a;
            }

            if (b.ToLower ().StartsWith ("the ") && b.Length > 4) {
                b2 = b.Substring (4).Trim ();
            } else {
                b2 = b;
            }

            return String.Compare (a2, b2);
        }

        private int OnTreeSort (TreeModel model, TreeIter a, TreeIter b) {
            Node nodea = (Node) store.GetValue (a, 0);
            Node nodeb = (Node) store.GetValue (b, 0);

            if (nodea is SongNode &&
                nodeb is SongNode) {
                return (nodea as SongNode).Song.TrackNumber.CompareTo ((nodeb as SongNode).Song.TrackNumber);
            } else if (nodea.Type == NodeType.Artist) {
                return ArtistCompare (nodea.Title, nodeb.Title);
            } else {
                return String.Compare (nodea.Title, nodeb.Title, true);
            }
        }

        private Song[] GetSongs (TreeIter parent) {
            TreeIter child = TreeIter.Zero;

            if (!store.IterChildren (out child, parent))
                return new Song[0];

            ArrayList songs = new ArrayList ();

            do {
                Node node = (Node) store.GetValue (child, 0);

                if (node is SongNode) {
                    songs.Add ((node as SongNode).Song);
                } else {
                    songs.AddRange (GetSongs (child));
                }
            } while (store.IterNext (ref child));

            return (Song[]) songs.ToArray (typeof (Song));
        }

        private void OnEdit (object o, EditedArgs args) {
            TreeIter iter = TreeIter.Zero;

            if (!store.GetIter (out iter, new TreePath (args.Path)))
                return;

            Node node = (Node) store.GetValue (iter, 0);
            node.Title = args.NewText;

            if (node is SongNode && SongEdited != null) {
                SongEdited (this, (node as SongNode).Song);
            }

            Song[] songs = GetSongs (iter);
            foreach (Song song in songs) {
                switch (node.Type) {
                case NodeType.Artist:
                    song.Artist = args.NewText;
                    break;
                case NodeType.Album:
                    song.Album = args.NewText;
                    break;
                default:
                    break;
                }

                if (SongEdited != null)
                    SongEdited (this, song);
            }
        }

        private void OnCellData (TreeViewColumn column, CellRenderer renderer, TreeModel model,
                                 TreeIter iter) {
            Node node = (Node) store.GetValue (iter, 0);

            if (node.Marked) {
                (renderer as CellRendererText).Markup = "<b>" + GLib.Markup.EscapeText (node.Title) + "</b>";
            } else {
                (renderer as CellRendererText).Text = node.Title;
            }
        }

        private TreeIter FindSong (Song song) {
            TreeIter albumIter = FindAlbum (song.IsCompilation ? Catalog.GetString ("Compilation") : song.Artist,
                                            song.Album);
            TreeIter iter = TreeIter.Zero;

            if (!store.IterChildren (out iter, albumIter)) {
                return TreeIter.Zero;
            }
            
            do {
                SongNode node = (SongNode) store.GetValue (iter, 0);

                if (node.Song == song)
                    return iter;
            } while (store.IterNext (ref iter));

            return TreeIter.Zero;
        }

        private TreeIter FindAlbum (string artist, string album) {

            TreeIter artistIter = FindArtist (artist);
            TreeIter iter = TreeIter.Zero;

            if (!store.IterChildren (out iter, artistIter)) {
                return store.AppendValues (artistIter, new GenericNode (NodeType.Album, album));
            }

            do {
                Node node = (Node) store.GetValue (iter, 0);

                if (node.Title == album) {
                    return iter;
                }
            } while (store.IterNext (ref iter));

            return store.AppendValues (artistIter, new GenericNode (NodeType.Album, album));
        }
        
        private TreeIter FindArtist (string artist) {

            TreeIter iter = TreeIter.Zero;

            store.GetIterFirst (out iter);

            if (iter.Equals (TreeIter.Zero)) {
                return store.AppendValues (new GenericNode (NodeType.Artist, artist));
            }

            do {
                Node node = (Node) store.GetValue (iter, 0);
                if (node.Title == artist) {
                    return iter;
                }

            } while (store.IterNext (ref iter));


            return store.AppendValues (new GenericNode (NodeType.Artist, artist));
        }

        private void Mark (Song song) {
            TreeIter iter = FindSong (song);

            do {
                Node node = (Node) store.GetValue (iter, 0);

                if (node.Marked == true)
                    return;
                
                node.Marked = true;

                store.EmitRowChanged (store.GetPath (iter), iter);
            } while (store.IterParent (out iter, iter));
        }

        private bool IsChildMarked (TreeIter iter) {
            TreeIter child = TreeIter.Zero;

            if (!store.IterChildren (out child, iter))
                return false;

            do {
                Node node = (Node) store.GetValue (child, 0);
                if (node.Marked)
                    return true;
                
            } while (store.IterNext (ref child));

            return false;
        }

        private void Unmark (Song song) {
            TreeIter iter = FindSong (song);

            do {
                Node node = (Node) store.GetValue (iter, 0);

                if (node.Marked == false ||
                    node.Type == NodeType.Artist)
                    return;
                
                node.Marked = false;
                store.EmitRowChanged (store.GetPath (iter), iter);

                if (IsChildMarked (iter))
                    break;
            } while (store.IterParent (out iter, iter));
        }

        private void OnNotify () {
            Refresh ();
        }

        private void OnSaveEnded (object o, EventArgs args) {
            notify.WakeupMain ();
        }

        public void AddDirectory (string directory) {
            AddDirectoryHelper helper = new AddDirectoryHelper (this, directory);
            helper.Run ();
        }

        public void AddSong (string file) {
            AddSong (file, true);
        }

        public bool IsPlayable (string file) {
            switch (System.IO.Path.GetExtension (file.ToLower ())) {
            case ".mp3":
            case ".m4p":
            case ".m4a":
                return true;
            default:
                return false;
            }
        }
        
        private Song GetSong (string file) {
            if (!IsPlayable (file)) {
                throw new ApplicationException (String.Format (Catalog.GetString ("Cannot add file '{0}', because the iPod is not able play this type of file."), file));
            }

            AudioFileWrapper afw = new AudioFileWrapper (file);
            
            Song song = db.CreateSong ();
            song.Artist = afw.Artist;
            song.Album = afw.Album;
            song.Title = afw.Title;
            song.Duration = TimeSpan.FromSeconds (afw.Duration);
            song.Year = afw.Year;
            song.TrackNumber = afw.TrackNumber;
            song.BitRate = afw.Bitrate;
            song.Genre = afw.Genre;
            song.FileName = file;
            
            if (song.Artist == null)
                song.Artist = String.Empty;
            
            if (song.Album == null)
                song.Album = String.Empty;
            
            if (song.Title == null)
                song.Title = Catalog.GetString ("Unknown");

            if (song.Genre == null)
                song.Genre = String.Empty;

            return song;
        }
        
        private void AddSong (string file, bool mark) {
            Song song = GetSong (file);
            Console.WriteLine ("Adding song: " + song.Title);
            AddSong (song, mark);
        }
        
        private void AddSong (Song song, bool mark) {
            SongNode node = new SongNode (song);
            
            store.AppendValues (FindAlbum (song.IsCompilation ? Catalog.GetString ("Compilation") : song.Artist,
                                           song.Album), node);

            if (mark) {
                Mark (song);
            }
        }

        public void Refresh () {
            if (db != null) {
                store.Clear ();
                
                foreach (Song song in db.Songs) {
                    AddSong (song, false);
                }
            } else {
                store.Clear ();
            }
        }

        private class AddDirectoryHelper {

            private static int numhelpers = 0;
            
            private SongView view;
            private string dir;
            private Thread thread;
            private ArrayList songs = new ArrayList ();
            private ThreadNotify notify;
            
            public AddDirectoryHelper (SongView view, string dir) {
                this.view = view;
                this.dir = dir;

                notify = new ThreadNotify (new ReadyEvent (OnNotify));
                thread = new Thread (new ThreadStart (ScanDirectory));
            }

            private void OnNotify () {
                foreach (Song song in songs) {
                    view.AddSong (song, true);
                }

                lock (typeof (AddDirectoryHelper)) {
                    numhelpers--;

                    if (numhelpers == 0) {
                        view.window.IsBusy = false;
                        view.window.SetSensitive (true);
                    }
                }
            }

            private void ScanDirectory () {
                try {
                    ScanDirectory (dir);
                } catch (Exception e) {
                    numhelpers--;
                } finally {
                    notify.WakeupMain ();
                }
            }

            private void ScanDirectory (string directory) {
                foreach (string subdir in Directory.GetDirectories (directory)) {
                    ScanDirectory (subdir);
                }
                
                foreach (string file in Directory.GetFiles (directory)) {
                    if (view.IsPlayable (file)) {
                        try {
                            songs.Add (view.GetSong (file));
                        } catch (Exception e) {
                            Console.WriteLine ("Failed to add song '{0}': {1}", file, e);
                        }
                    }
                }
            }

            public void Run () {
                lock (typeof (AddDirectoryHelper)) {
                    numhelpers++;
                }
                
                view.window.IsBusy = true;
                view.window.SetSensitive (false);
                thread.Start ();
            }
        }
    }
}
